package io.gitlab.jfronny.muscript.data.dynamic;

import io.gitlab.jfronny.commons.StringFormatter;
import io.gitlab.jfronny.muscript.StandardLib;
import io.gitlab.jfronny.muscript.ast.DynamicExpr;
import io.gitlab.jfronny.muscript.compiler.ExprWriter;
import io.gitlab.jfronny.muscript.compiler.Parser;
import io.gitlab.jfronny.muscript.data.dynamic.additional.*;

import java.io.IOException;

/**
 * Represents a value of an unknown type
 * Override toString(StringBuilder) to support custom serialization (note: the serialized form is ran with muScript to generate the tree)
 * @param <T> the type represented
 */
public sealed interface Dynamic<T> permits DBool, DNumber, DString, DObject, DList, DCallable, DNull, DContainer {
    /**
     * Deserialize simply dynamic values. DOES NOT SUPPORT CALLABLES!
     * Use Parser.parse() if you truly need to deserialize them, but be aware that that might enable DOS attacks
     */
    static Dynamic<?> deserialize(String source) {
        DynamicExpr expr = Parser.parse(source).asDynamicExpr();
        if (!DirectPreconditionVisitor.visit(expr)) throw new IllegalArgumentException("This expression does not directly express a dynamic and may not be loaded this way");
        return expr.get(StandardLib.createScope());
    }

    static String serialize(Dynamic<?> dynamic) {
        StringBuilder sb = new StringBuilder();
        try (ExprWriter ew = new ExprWriter(sb, true)) {
            dynamic.serialize(ew);
        } catch (IOException e) {
            throw new RuntimeException("Could not stringify", e);
        }
        return sb.toString();
    }

    void serialize(ExprWriter writer) throws IOException;
    T getValue();

    default DBool asBool() {
        if (this instanceof DBool bool) return bool;
        else throw new DynamicTypeConversionException("bool", this);
    }

    default DNumber asNumber() {
        if (this instanceof DNumber number) return number;
        else throw new DynamicTypeConversionException("number", this);
    }

    default DString asString() {
        if (this instanceof DString string) return string;
        else return DFinal.of(StringFormatter.toString(getValue()));
    }

    default DObject asObject() {
        if (this instanceof DObject object) return object;
        else throw new DynamicTypeConversionException("object", this);
    }

    default DList asList() {
        if (this instanceof DList list) return list;
        else throw new DynamicTypeConversionException("list", this);
    }

    default DCallable asCallable() {
        if (this instanceof DCallable callable) return callable;
        else throw new DynamicTypeConversionException("callable", this);
    }
}
